@@ -51,6 +51,7 @@
 
 #include <config.h>
 
+#include <libgnomevfs/gnome-vfs-i18n.h>
 #include <libgnomevfs/gnome-vfs-context.h>
 #include <libgnomevfs/gnome-vfs-method.h>
 #include <libgnomevfs/gnome-vfs-module.h>
@@ -81,6 +82,7 @@ typedef struct {
 	gchar         *hash_name;
 	gint           in_fd;
 	gint           out_fd;
+	gint           tty_fd;
 	GIOChannel    *error_channel;
 	pid_t          ssh_pid;
 
@@ -180,11 +182,14 @@ atomic_io (read_write_fn f, gint fd, gpointer buffer_in, gsize size)
 {
 	gsize pos = 0, res;
 	guchar *buffer;
+        long int __result;
 
 	buffer = buffer_in;
 
 	while (pos < size) {
-		res = TEMP_FAILURE_RETRY (f (fd, buffer, size - pos));
+	  	do __result = (long int) (f (fd, buffer, size - pos));
+		while (__result == -1L && errno == EINTR);
+		res = __result;
 
 		if (res < 0)
 			return -1;
@@ -511,13 +516,14 @@ buffer_read_file_info (Buffer *buf, GnomeVFSFileInfo *info)
 			       info->permissions));
 
 		info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_TYPE;
-		if (info->permissions & S_IFREG) info->type = GNOME_VFS_FILE_TYPE_REGULAR;
-		else if (info->permissions & S_IFDIR) info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
-		else if (info->permissions & S_IFIFO) info->type = GNOME_VFS_FILE_TYPE_FIFO;
-		else if (info->permissions & S_IFSOCK) info->type = GNOME_VFS_FILE_TYPE_SOCKET;
-		else if (info->permissions & S_IFCHR) info->type = GNOME_VFS_FILE_TYPE_CHARACTER_DEVICE;
-		else if (info->permissions & S_IFBLK) info->type = GNOME_VFS_FILE_TYPE_BLOCK_DEVICE;
-		else if (info->permissions & S_IFLNK) info->type = GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK;
+		
+		if (S_ISREG (info->permissions)) info->type = GNOME_VFS_FILE_TYPE_REGULAR;		
+		else if (S_ISDIR (info->permissions)) info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
+		else if (S_ISFIFO (info->permissions)) info->type = GNOME_VFS_FILE_TYPE_FIFO;
+		else if (S_ISSOCK (info->permissions)) info->type = GNOME_VFS_FILE_TYPE_SOCKET;
+		else if (S_ISCHR (info->permissions)) info->type = GNOME_VFS_FILE_TYPE_CHARACTER_DEVICE;
+		else if (S_ISBLK (info->permissions)) info->type = GNOME_VFS_FILE_TYPE_BLOCK_DEVICE;
+		else if (S_ISLNK (info->permissions)) info->type = GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK;
 		else info->type = GNOME_VFS_FILE_TYPE_UNKNOWN;
 
 		info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_FLAGS;
@@ -840,24 +846,31 @@ iobuf_send_string_request_with_file_info (int                      fd,
 
 
 
-static gboolean
-invoke_full_auth (const GnomeVFSURI *uri,
-		  gboolean done_auth,
-		  const char *password_line,
-		  char **password_out)
+static char*
+get_user_from_uri_or_password_line (const GnomeVFSURI *uri,
+				    const char *password_line)
 {
-	GnomeVFSModuleCallbackFullAuthenticationIn in_args;
-	GnomeVFSModuleCallbackFullAuthenticationOut out_args;
-	gboolean invoked;
-	char *user, *chr, *ptr;
-	char *object;
-	gboolean passphrase;
+	char *chr, *user = NULL;
 
-	passphrase = g_str_has_prefix (password_line, "Enter passphrase for key");
+	if (!g_str_has_prefix (password_line, "Enter passphrase for key")) {
+		chr = strchr (password_line, '@');
 
-	user = NULL;
-	object = NULL;
-	if (passphrase) {
+		if (chr != NULL) {
+			user = g_strndup (password_line, chr - password_line);
+		}
+	}
+	if (user == NULL) {
+		user = g_strdup ((char *)gnome_vfs_uri_get_user_name (uri));
+	}
+	return user;
+}
+
+static char*
+get_object_from_password_line (const char *password_line)
+{
+	char *chr, *ptr, *object = NULL;
+
+	if (g_str_has_prefix (password_line, "Enter passphrase for key")) {
 		ptr = strchr (password_line, '\'');
 		if (ptr != NULL) {
 			ptr += 1;
@@ -868,13 +881,89 @@ invoke_full_auth (const GnomeVFSURI *uri,
 				object = g_strdup (ptr);
 			}
 		}
-	} else {
-		chr = strchr (password_line, '@');
-
-		if (chr != NULL) {
-			user = g_strndup (password_line, chr - password_line);
-		}
 	}
+	return object;
+}
+
+static char*
+get_server_from_uri_or_password_line (const GnomeVFSURI *uri,
+				      const char *password_line)
+{
+	if (!g_str_has_prefix (password_line, "Enter passphrase for key")) {
+		return g_strdup ( (char *)gnome_vfs_uri_get_host_name (uri));
+	}
+	return NULL;
+}
+
+static char*
+get_authtype_from_password_line (const char *password_line)
+{
+	if (g_str_has_prefix (password_line, "Enter passphrase for key")) {
+		return g_strdup ("publickey");
+	} 
+	return g_strdup ("password");
+}
+
+
+
+static gboolean
+invoke_fill_auth (const GnomeVFSURI *uri,
+		  const char *password_line,
+		  char **password_out)
+{
+	gboolean invoked;
+	GnomeVFSModuleCallbackFillAuthenticationIn in_args;
+	GnomeVFSModuleCallbackFillAuthenticationOut out_args;
+
+
+	memset (&in_args, 0, sizeof (in_args));
+	in_args.protocol = "sftp";
+	
+	in_args.uri = gnome_vfs_uri_to_string (uri, 0);
+	in_args.object = get_object_from_password_line (password_line);
+	in_args.authtype = get_authtype_from_password_line (password_line);
+	in_args.domain = NULL;
+	in_args.port = gnome_vfs_uri_get_host_port (uri);
+	in_args.server = get_server_from_uri_or_password_line (uri, password_line);
+	in_args.username = get_user_from_uri_or_password_line (uri, password_line);
+	memset (&out_args, 0, sizeof (out_args));
+
+	invoked = gnome_vfs_module_callback_invoke
+		  (GNOME_VFS_MODULE_CALLBACK_FILL_AUTHENTICATION,
+		  &in_args, sizeof (in_args),
+		  &out_args, sizeof (out_args));
+	if (invoked && out_args.valid) {
+		*password_out = g_strdup (out_args.password);
+		g_free (out_args.username);
+		g_free (out_args.domain);
+		g_free (out_args.password);
+	} else {
+		*password_out = NULL;
+	}
+
+	g_free (in_args.uri);
+	g_free (in_args.username);
+	g_free (in_args.object);
+	g_free (in_args.server);
+	g_free (in_args.authtype);
+
+	return invoked && out_args.valid;
+}
+
+static gboolean
+invoke_full_auth (const GnomeVFSURI *uri,
+		  gboolean done_auth,
+		  const char *password_line,
+		  char **password_out,
+		  char **keyring_out,
+		  char **user_out,
+		  char **object_out,
+		  char **authtype_out,
+		  gboolean *save_password_out)
+{
+	GnomeVFSModuleCallbackFullAuthenticationIn in_args;
+	GnomeVFSModuleCallbackFullAuthenticationOut out_args;
+	gboolean invoked;
 
 	memset (&in_args, 0, sizeof (in_args));
 	in_args.uri = gnome_vfs_uri_to_string (uri, 0);
@@ -882,22 +971,13 @@ invoke_full_auth (const GnomeVFSURI *uri,
 	if (done_auth) {
 		in_args.flags |= GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_PREVIOUS_ATTEMPT_FAILED;
 	}
-	/* TODO: Support username too. A bit harder */
-
 	in_args.protocol = "sftp";
-	if (!passphrase) {
-		in_args.server = (char *)gnome_vfs_uri_get_host_name (uri);
-		if (user != NULL) {
-			in_args.username = user;
-		} else {
-			in_args.username = (char *)gnome_vfs_uri_get_user_name (uri);
-		}
-	}
-	in_args.object = object;
-	in_args.authtype = passphrase ? "publickey": "password";
-	
+	in_args.object = get_object_from_password_line (password_line);
+	in_args.authtype = get_authtype_from_password_line (password_line);
 	in_args.domain = NULL;
 	in_args.port = gnome_vfs_uri_get_host_port (uri);
+	in_args.server = get_server_from_uri_or_password_line (uri, password_line);
+	in_args.username = get_user_from_uri_or_password_line (uri, password_line);
 
 	memset (&out_args, 0, sizeof (out_args));
 
@@ -905,24 +985,99 @@ invoke_full_auth (const GnomeVFSURI *uri,
 		(GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION,
 		 &in_args, sizeof (in_args),
 		 &out_args, sizeof (out_args));
-
 	if (invoked && !out_args.abort_auth) {
+		if (out_args.save_password) {
+			*keyring_out = g_strdup (out_args.keyring);
+			*user_out = get_user_from_uri_or_password_line (uri, password_line);
+			*object_out = get_object_from_password_line (password_line);
+			*authtype_out = get_authtype_from_password_line (password_line);
+		}
 		*password_out = g_strdup (out_args.password);
+		*save_password_out = out_args.save_password;
+		g_free (out_args.username);
+		g_free (out_args.domain);
+		g_free (out_args.password);
+		g_free (out_args.keyring);
 	} else {
 		*password_out = NULL;
 	}
-	g_free (out_args.username);
-	g_free (out_args.domain);
-	g_free (out_args.password);
-	g_free (out_args.keyring);
 
 	g_free (in_args.uri);
-	g_free (user);
-	g_free (object);
+	g_free (in_args.username);
+	g_free (in_args.object);
+	g_free (in_args.server);
+	g_free (in_args.authtype);
 
 	return invoked && !out_args.abort_auth;
 }
 
+static void
+invoke_save_auth (const GnomeVFSURI *uri,
+	   char *keyring,
+	   char *user,
+	   char *object,
+	   char *authtype,
+	   char *password)
+{
+	GnomeVFSModuleCallbackSaveAuthenticationIn save_in_args;
+	GnomeVFSModuleCallbackSaveAuthenticationOut save_out_args;
+
+	memset (&save_in_args, 0, sizeof (save_in_args));
+	save_in_args.uri = gnome_vfs_uri_to_string (uri, 0);
+	save_in_args.server = (char *)gnome_vfs_uri_get_host_name (uri);
+	save_in_args.port = gnome_vfs_uri_get_host_port (uri);
+	save_in_args.protocol = "sftp";
+	save_in_args.keyring = keyring;
+	save_in_args.username = user;
+	save_in_args.object = object;
+	save_in_args.authtype = authtype;
+	save_in_args.password = password;
+
+	memset (&save_out_args, 0, sizeof (save_out_args));
+	gnome_vfs_module_callback_invoke (GNOME_VFS_MODULE_CALLBACK_SAVE_AUTHENTICATION,
+					  &save_in_args, sizeof (save_in_args),
+					  &save_out_args, sizeof (save_out_args));
+	g_free (save_in_args.uri);
+}
+	
+
+typedef enum {
+	SFTP_VENDOR_INVALID = 0,
+	SFTP_VENDOR_OPENSSH,
+	SFTP_VENDOR_SSH
+} SFTPClientVendor;
+
+static SFTPClientVendor
+get_sftp_client_vendor (void)
+{
+	char *ssh_stderr;
+	char *args[3];
+	gint ssh_exitcode;
+	SFTPClientVendor res = SFTP_VENDOR_INVALID;
+	
+	args[0] = g_strdup (SSH_PROGRAM);
+	args[1] = g_strdup ("-V");
+	args[2] = NULL;
+	if (g_spawn_sync (NULL, args, NULL,
+			  G_SPAWN_SEARCH_PATH | G_SPAWN_STDOUT_TO_DEV_NULL,
+			  NULL, NULL,
+			  NULL, &ssh_stderr,
+			  &ssh_exitcode, NULL)) {
+		if (ssh_stderr == NULL)
+			res = SFTP_VENDOR_INVALID;
+		else if (strstr (ssh_stderr, "OpenSSH") != NULL)
+			res = SFTP_VENDOR_OPENSSH;
+		else if (strstr (ssh_stderr, "SSH Secure Shell") != NULL)
+			res = SFTP_VENDOR_SSH;
+		else
+			res = SFTP_VENDOR_INVALID;
+	}
+	g_free (ssh_stderr);
+	g_free (args[0]);
+	g_free (args[1]);
+
+	return res;
+}
 
 /* Derived from OpenSSH, sftp.c:main */
 
@@ -930,34 +1085,60 @@ static GnomeVFSResult
 sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri) 
 {
 	GnomeVFSResult  res;
-	GIOChannel     *error_channel;
-	GIOChannel     *tty_channel;
+	GIOChannel     *error_channel = NULL;
+	GIOChannel     *tty_channel = NULL;
 	int             in_fd, out_fd, err_fd, tty_fd;
 	pid_t           ssh_pid;
 	const gchar    *user_name;
 	gint            port;
 	guint           last_arg, i;
+	gboolean        full_auth;
 	gboolean        done_auth;
+	gboolean	save_password;
 	Buffer          msg;
 	gchar           type;
+	char *password = NULL;
+	char *keyring  = NULL;
+	char *user     = NULL;
+	char *object   = NULL;
+	char *authtype = NULL;
 
 	GError         *error = NULL;
 
 	gchar          *args[20]; /* Enough for now, extend if you add more args */
+	gboolean invoked; 		  
+	GnomeVFSModuleCallbackQuestionIn in_args; 
+	GnomeVFSModuleCallbackQuestionOut out_args;
 
+	SFTPClientVendor client_vendor;
+	
 	DEBUG (gchar *tmp);
 
+	client_vendor = get_sftp_client_vendor ();
+	
 	/* Fill in the first few args */
 	last_arg = 0;
 	args[last_arg++] = g_strdup (SSH_PROGRAM);
-	args[last_arg++] = g_strdup ("-oForwardX11 no");
-	args[last_arg++] = g_strdup ("-oForwardAgent no");
-	args[last_arg++] = g_strdup ("-oClearAllForwardings yes");
-	args[last_arg++] = g_strdup ("-oProtocol 2");
+
+	if (client_vendor == SFTP_VENDOR_OPENSSH) {
+		args[last_arg++] = g_strdup ("-oForwardX11 no");
+		args[last_arg++] = g_strdup ("-oForwardAgent no");
+		args[last_arg++] = g_strdup ("-oClearAllForwardings yes");
+		args[last_arg++] = g_strdup ("-oProtocol 2");
+		args[last_arg++] = g_strdup ("-oNoHostAuthenticationForLocalhost yes");
 #ifndef USE_PTY
-	args[last_arg++] = g_strdup ("-oBatchMode yes");
+		args[last_arg++] = g_strdup ("-oBatchMode yes");
 #endif
-	args[last_arg++] = g_strdup ("-oNoHostAuthenticationForLocalhost yes");
+
+	} else if (client_vendor == SFTP_VENDOR_SSH) {
+		args[last_arg++] = g_strdup ("-x");
+	} else {
+		for (i = 0; i < last_arg; i++) {
+			g_free (args[i]);
+		}
+		return GNOME_VFS_ERROR_INTERNAL;
+	}
+
 
 	/* Disable login prompt for now */
 	/* args[last_arg++] = g_strdup ("-oBatchMode yes"); */
@@ -981,8 +1162,13 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 
 	args[last_arg++] = g_strdup ("-s");
 
-	args[last_arg++] = g_strdup (gnome_vfs_uri_get_host_name (uri));
-	args[last_arg++] = g_strdup ("sftp");
+	if (client_vendor == SFTP_VENDOR_SSH) {
+		args[last_arg++] = g_strdup ("sftp");
+		args[last_arg++] = g_strdup (gnome_vfs_uri_get_host_name (uri));
+	} else {
+		args[last_arg++] = g_strdup (gnome_vfs_uri_get_host_name (uri));
+		args[last_arg++] = g_strdup ("sftp");
+	}
 
 	args[last_arg++] = NULL;
 
@@ -1038,42 +1224,71 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 					g_io_channel_get_flags (tty_channel) | G_IO_FLAG_NONBLOCK, NULL);
 	}
 	done_auth = FALSE;
+	full_auth = FALSE;
 	while (tty_fd != -1) {
 		fd_set ifds;
 		struct timeval tv;
 		int ret;
+		int prompt_fd;
+		GIOChannel *prompt_channel;
 		GIOStatus io_status;
 		char buffer[1024];
 		gsize len;
-		char *password;
+		char *choices[3];
+		char *pos;
+		char *startpos;
+		char *endpos;
+		char *hostname = NULL;
+		char *fingerprint = NULL;
 
+		if (client_vendor == SFTP_VENDOR_SSH) {
+			prompt_fd = err_fd;
+			prompt_channel = error_channel;
+		} else {
+			prompt_fd = tty_fd;
+			prompt_channel = tty_channel;
+		}
+		
 		FD_ZERO (&ifds);
 		FD_SET (in_fd, &ifds);
-		FD_SET (tty_fd, &ifds);
-		
+		FD_SET (prompt_fd, &ifds);
+
 		tv.tv_sec = 10;
 		tv.tv_usec = 0;
-		
-		ret = select (MAX (in_fd, tty_fd)+1, &ifds, NULL, NULL, &tv);
-		if (ret == 0) {
+
+		ret = select (MAX (in_fd, prompt_fd)+1, &ifds, NULL, NULL, &tv);
+
+		if (ret <= 0) {
 			/* Timeout */
 			res = GNOME_VFS_ERROR_IO;
 			goto bail;
 		}
-		
+
 		if (FD_ISSET (in_fd, &ifds)) {
 			break;
 		}
 
-		g_assert (FD_ISSET (tty_fd, &ifds));
+		g_assert (FD_ISSET (prompt_fd, &ifds));
 
 		error = NULL;
-		io_status = g_io_channel_read_chars (tty_channel, buffer, sizeof(buffer)-1, &len, &error);
+		io_status = g_io_channel_read_chars (prompt_channel, buffer, sizeof(buffer)-1, &len, &error);
 		if (io_status == G_IO_STATUS_NORMAL) {
 			buffer[len] = 0;
 			if (g_str_has_suffix (buffer, "password: ") ||
+			    g_str_has_suffix (buffer, "Password: ") ||
 			    g_str_has_prefix (buffer, "Enter passphrase for key")) {
-				if (invoke_full_auth (uri, done_auth, buffer, &password) && password != NULL) {
+				if (!done_auth && gnome_vfs_uri_get_password (uri) != NULL) {
+					password = g_strdup (gnome_vfs_uri_get_password (uri));
+					g_io_channel_write_chars (tty_channel, password, -1, &len, NULL);
+					g_io_channel_write_chars (tty_channel, "\n", 1, &len, NULL);
+					g_io_channel_flush (tty_channel, NULL);
+				} else if (!done_auth && invoke_fill_auth (uri, buffer, &password) && password != NULL) {
+					g_io_channel_write_chars (tty_channel, password, -1, &len, NULL);
+					g_io_channel_write_chars (tty_channel, "\n", 1, &len, NULL);
+					g_io_channel_flush (tty_channel, NULL);
+				} else if (invoke_full_auth (uri, done_auth, buffer, &password, &keyring, 
+							     &user, &object, &authtype, &save_password) && password != NULL) {
+					full_auth = TRUE;
 					g_io_channel_write_chars (tty_channel, password, -1, &len, NULL);
 					g_io_channel_write_chars (tty_channel, "\n", 1, &len, NULL);
 					g_io_channel_flush (tty_channel, NULL);
@@ -1082,12 +1297,95 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 					goto bail;
 				}
 				done_auth = TRUE;
-			} else if (g_str_has_prefix (buffer, "The authenticity of host")) {
-				/* FIXME: This should do a callback asking the user if the host id is ok */
-				/* For now we just fail */
-				res = GNOME_VFS_ERROR_ACCESS_DENIED;
-				goto bail;
-                        }
+			} else if (g_str_has_prefix (buffer, "The authenticity of host '") ||
+				   strstr (buffer, "Key fingerprint:") != NULL) {
+
+				if (g_str_has_prefix (buffer, "The authenticity of host '")) {
+					/* OpenSSH */
+					pos = strchr (&buffer[26], '\'');
+					if (pos == NULL) {
+						res = GNOME_VFS_ERROR_GENERIC;
+						goto bail;
+					}
+
+
+					hostname = g_strndup (&buffer[26], pos - (&buffer[26]));
+
+					startpos = strstr (pos, " key fingerprint is ");
+					if (startpos == NULL) {
+						res = GNOME_VFS_ERROR_GENERIC;
+						g_free (hostname);
+						goto bail;
+					}
+
+					startpos = startpos + 20;
+					endpos = strchr (startpos, '.');
+					if (endpos == NULL) {
+						res = GNOME_VFS_ERROR_GENERIC;
+						g_free (hostname);
+						goto bail;
+					}
+					fingerprint = g_strndup (startpos, endpos - startpos);
+				} else if (strstr (buffer, "Key fingerprint:") != NULL) {
+					/* SSH.com*/
+					hostname = g_strdup (gnome_vfs_uri_get_host_name (uri));
+					startpos = strstr (buffer, "Key fingerprint:");
+					if (startpos == NULL) {
+						g_free (hostname);
+						res = GNOME_VFS_ERROR_GENERIC;
+						goto bail;
+					}
+					startpos = startpos + 18;
+					endpos = strchr (startpos, '\r');
+					fingerprint = g_strndup (startpos, endpos - startpos);
+				} else {
+					res = GNOME_VFS_ERROR_GENERIC;
+					goto bail;	
+				}
+				
+				in_args.primary_message = g_strdup_printf (_("The identity of the remote computer (%s) is unknown."), hostname);
+				in_args.secondary_message = g_strdup_printf (_("This happens when you log in to a computer the first time.\n\n"
+									       "The identity sent by the remote computer is %s. "
+									       "If you want to be absolutely sure it is safe to continue, contact the system administrator."), fingerprint);
+
+				g_free (hostname);
+				g_free (fingerprint);
+
+				in_args.choices = choices;
+				in_args.choices[0] = _("Log In Anyway");
+				in_args.choices[1] = _("Cancel Login");
+				in_args.choices[2] = NULL;
+								
+				invoked = gnome_vfs_module_callback_invoke
+					(GNOME_VFS_MODULE_CALLBACK_QUESTION,
+					 &in_args, sizeof (in_args),
+					 &out_args, sizeof (out_args));
+				
+				if (invoked) {
+					if (out_args.answer == 0) {
+						g_io_channel_write_chars (tty_channel, "yes\n", -1, &len, NULL);
+					} else {
+						g_io_channel_write_chars (tty_channel, "no\n", -1, &len, NULL);
+						g_free (in_args.primary_message);
+						g_free (in_args.secondary_message);
+						res = GNOME_VFS_ERROR_ACCESS_DENIED;
+						goto bail;
+					}
+					g_io_channel_flush (tty_channel, NULL);
+					buffer[0]='\0';
+				} else {
+					g_io_channel_write_chars (tty_channel, "no\n", -1, &len, NULL);
+					g_io_channel_flush (tty_channel, NULL);
+					buffer[0]='\0';
+					g_free (in_args.primary_message);
+					g_free (in_args.secondary_message);
+					res = GNOME_VFS_ERROR_ACCESS_DENIED;
+					goto bail;
+				}
+				g_free (in_args.primary_message);
+				g_free (in_args.secondary_message);
+			}
+			
 		}
 	}
 
@@ -1108,6 +1406,9 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 		res = GNOME_VFS_ERROR_PROTOCOL_ERROR;
 	} else {
 		/* Everything's A-OK. Set up the connection and go */
+		if (full_auth == TRUE && save_password == TRUE) {
+			invoke_save_auth (uri, keyring, user, object, authtype, password);
+		}
 
 		if (!g_thread_supported ()) g_thread_init (NULL);
 
@@ -1115,7 +1416,9 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 		(*connection)->ref_count = 1;
 		(*connection)->in_fd = in_fd;
 		(*connection)->out_fd = out_fd;
+		(*connection)->tty_fd = tty_fd;
 		(*connection)->error_channel = error_channel;
+		g_io_channel_ref (error_channel);
 		(*connection)->ssh_pid = ssh_pid;
 		(*connection)->version = buffer_read_gint32 (&msg);
 		(*connection)->mutex = g_mutex_new ();
@@ -1131,12 +1434,25 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 	}
 
  bail:
-
 	buffer_free (&msg);
 
+	g_free (password);
+	g_free (keyring);
+	g_free (user);
+	g_free (object);
+	g_free (authtype);
+	if (error_channel != NULL) {
+		g_io_channel_unref (error_channel);
+	}
+	if (tty_channel != NULL) {
+		g_io_channel_unref (tty_channel);
+	}
 	if (res != GNOME_VFS_OK) {
 		close (in_fd);
 		close (out_fd);
+		close (err_fd);
+		if (tty_fd != -1)
+			close (tty_fd);
 		*connection = NULL;
 
 		/* TODO: Do we leak error_channel and connection? */
@@ -1168,6 +1484,12 @@ sftp_get_connection (SftpConnection **connection, const GnomeVFSURI *uri)
 	if (user_name == NULL)
 		user_name = g_get_user_name ();
 
+	if (host_name == NULL)
+	{
+		res = GNOME_VFS_ERROR_HOST_NOT_FOUND;
+		goto bail;
+	}
+
 	hash_name = g_strconcat (user_name, "@", host_name, NULL);
 
 	DEBUG (g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG,
@@ -1182,7 +1504,12 @@ sftp_get_connection (SftpConnection **connection, const GnomeVFSURI *uri)
 		res = sftp_connect (connection, uri);
 
 		if (res == GNOME_VFS_OK) {
-			g_return_val_if_fail (*connection != NULL, GNOME_VFS_ERROR_INTERNAL);
+			if (*connection == NULL)
+			{
+				res = GNOME_VFS_ERROR_INTERNAL;
+				g_free (hash_name);
+				goto bail;
+			}
 			g_mutex_lock ((*connection)->mutex);
 			(*connection)->hash_name = hash_name;
 			g_hash_table_insert (sftp_connection_table, hash_name, *connection);
@@ -1214,6 +1541,7 @@ sftp_get_connection (SftpConnection **connection, const GnomeVFSURI *uri)
 		res = GNOME_VFS_OK;
 	}
 
+ bail:
 	G_UNLOCK (sftp_connection_table);
 
 	return res;
@@ -1226,6 +1554,8 @@ sftp_connection_close (SftpConnection *conn)
 
 	close (conn->in_fd);
 	close (conn->out_fd);
+	if (conn->tty_fd != -1)
+		close (conn->tty_fd);
 	g_source_remove (conn->event_id);
 	g_io_channel_shutdown (conn->error_channel, FALSE, NULL);
 	g_io_channel_unref (conn->error_channel);
@@ -1484,7 +1814,7 @@ do_open (GnomeVFSMethod        *method,
 
 	buffer_write_gint32 (&msg, sftp_mode);
 
-	bzero (&info, sizeof (GnomeVFSFileInfo));
+	memset (&info, 0, sizeof (GnomeVFSFileInfo));
 	buffer_write_file_info (&msg, &info, GNOME_VFS_SET_FILE_INFO_NONE);
 
 	buffer_send (&msg, conn->out_fd);
@@ -1555,11 +1885,19 @@ do_create (GnomeVFSMethod        *method,
 	ssh_mode = SSH2_FXF_CREAT;
 	if (mode & GNOME_VFS_OPEN_READ) ssh_mode |= SSH2_FXF_READ;
 	if (mode & GNOME_VFS_OPEN_WRITE) ssh_mode |= SSH2_FXF_WRITE;
-	if (exclusive) ssh_mode |= SSH2_FXF_EXCL;
+	if (exclusive) {
+		ssh_mode |= SSH2_FXF_EXCL;
+	} else {
+		/* It might be ok to unconditionnally add this truncation flag,
+		 * but I'm not 100% sure that SSH2_FXF_EXCL takes precedence
+		 * over SSH2_FXF_TRUNC, so better be safe than sorry ;)
+		 */
+		ssh_mode |= SSH2_FXF_TRUNC;
+	}
 
 	buffer_write_gint32 (&msg, ssh_mode);
 
-	bzero (&info, sizeof (GnomeVFSFileInfo));
+	memset (&info, 0,  sizeof (GnomeVFSFileInfo));
 	info.permissions = perm;
 	buffer_write_file_info (&msg, &info, GNOME_VFS_SET_FILE_INFO_PERMISSIONS);
 
@@ -1632,7 +1970,7 @@ do_close (GnomeVFSMethod       *method,
 
 	DEBUG (g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "%s: Exit", __FUNCTION__));
 
-	return sftp_status_to_vfs_result (status);
+	return status;
 }
 
 static GnomeVFSResult 
@@ -1653,6 +1991,8 @@ do_read (GnomeVFSMethod       *method,
 	guchar *curr_ptr;
 	gboolean out_of_order;
 	GnomeVFSResult result;
+	gboolean got_eof;
+	guint outstanding;
 
 	struct ReadRequest 
 	{
@@ -1665,6 +2005,8 @@ do_read (GnomeVFSMethod       *method,
 
 	DEBUG (g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "%s: Enter", __FUNCTION__));
 
+	got_eof = FALSE;
+	
 	handle = SFTP_OPEN_HANDLE (method_handle);
 	*bytes_read = 0;
 	num_req = 0;
@@ -1672,6 +2014,7 @@ do_read (GnomeVFSMethod       *method,
 	req_svc_ptr = 0;
 	curr_ptr = buffer_in;
 	buffer = buffer_in;
+	outstanding = 0;
 
 	read_req = g_new0 (struct ReadRequest, max_req);
 
@@ -1679,7 +2022,8 @@ do_read (GnomeVFSMethod       *method,
 
 	sftp_connection_lock (handle->connection);
 
-	while (*bytes_read < num_bytes) {
+	while ((*bytes_read < num_bytes) ||
+	       (outstanding > 0)) {
 		while (num_req < max_req && curr_ptr < buffer + num_bytes) {
 			read_req[req_ptr].id = sftp_connection_get_id (handle->connection);
 			read_req[req_ptr].req_len =
@@ -1691,6 +2035,7 @@ do_read (GnomeVFSMethod       *method,
 				      __FUNCTION__, req_ptr, read_req[req_ptr].id,
 				      read_req[req_ptr].req_len, read_req[req_ptr].ptr));
 
+			outstanding++;
 			iobuf_send_read_request (handle->connection->out_fd,
 						 read_req[req_ptr].id,
 						 handle->offset + (read_req[req_ptr].ptr - buffer),
@@ -1706,6 +2051,7 @@ do_read (GnomeVFSMethod       *method,
 
 		buffer_clear (&msg);
 		result = buffer_recv (&msg, handle->connection->in_fd);
+		outstanding--;
 
 		if (result != GNOME_VFS_OK) {
 			sftp_connection_unlock (handle->connection);
@@ -1744,15 +2090,20 @@ do_read (GnomeVFSMethod       *method,
 			DEBUG (g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "%s: req_svc = %d, req_ptr = %d",
 				      __FUNCTION__, req_svc, req_ptr));
 
-			if (status != SSH2_FX_EOF || read_req[req_svc].ptr == buffer) {
+			if (status != SSH2_FX_EOF) {
+				DEBUG (g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "%s: status return %d",
+					      __FUNCTION__, sftp_status_to_vfs_result (status)));
 				buffer_free (&msg);
 				g_free (read_req);
 				sftp_connection_unlock (handle->connection);
 				return sftp_status_to_vfs_result (status);
 			}
 
+			if (read_req[req_svc].ptr == buffer)
+				got_eof = TRUE;
+			
 			/* We hit an EOF, so make sure we don't try any more reads after this */
-			num_bytes = read_req[req_svc].ptr - buffer;
+			num_bytes = MIN (num_bytes, read_req[req_svc].ptr - buffer);
 
 			/* Kill this request */
 			read_req[req_svc].id = 0;
@@ -1771,6 +2122,7 @@ do_read (GnomeVFSMethod       *method,
 				read_req[req_svc].req_len -= len;
 				read_req[req_svc].ptr += len;
 
+				outstanding++;
 				iobuf_send_read_request
 					(handle->connection->out_fd,
 					 read_req[req_svc].id,
@@ -1802,6 +2154,9 @@ do_read (GnomeVFSMethod       *method,
 
 	sftp_connection_unlock (handle->connection);
 
+	if (got_eof)
+		return GNOME_VFS_ERROR_EOF;
+	
 	return GNOME_VFS_OK;
 }
 
@@ -2255,9 +2610,9 @@ do_read_directory (GnomeVFSMethod       *method,
 			while (handle->info_write_ptr + count > handle->info_alloc) {
 				handle->info_alloc *= 2;
 				handle->info = g_renew (GnomeVFSFileInfo, handle->info, handle->info_alloc);
-				bzero (&(handle->info[handle->info_write_ptr]),
-				       sizeof (GnomeVFSFileInfo) *
-				       (handle->info_alloc - handle->info_write_ptr));
+				memset (&(handle->info[handle->info_write_ptr]), 0,
+					sizeof (GnomeVFSFileInfo) *
+					(handle->info_alloc - handle->info_write_ptr));
 			}
 		}
 
@@ -2335,14 +2690,14 @@ do_make_directory (GnomeVFSMethod  *method,
 	id = sftp_connection_get_id (conn);
 
 	path = gnome_vfs_unescape_string (gnome_vfs_uri_get_path (uri), NULL);
-	bzero (&info, sizeof (GnomeVFSFileInfo));
+	memset (&info, 0, sizeof (GnomeVFSFileInfo));
 	iobuf_send_string_request_with_file_info (conn->out_fd, id, SSH2_FXP_MKDIR,
 						  path, strlen (path), &info,
 						  GNOME_VFS_SET_FILE_INFO_NONE);
 
 	g_free (path);
 
-	res = sftp_status_to_vfs_result (iobuf_read_result (conn->in_fd, id));
+	res = iobuf_read_result (conn->in_fd, id);
 
 	sftp_connection_unref (conn);
 	sftp_connection_unlock (conn);
@@ -2371,7 +2726,7 @@ do_remove_directory (GnomeVFSMethod  *method,
 
 	g_free (path);
 
-	res = sftp_status_to_vfs_result (iobuf_read_result (conn->in_fd, id));
+	res = iobuf_read_result (conn->in_fd, id);
 
 	sftp_connection_unref (conn);
 	sftp_connection_unlock (conn);
@@ -2414,7 +2769,7 @@ do_move (GnomeVFSMethod  *method,
 	g_free (old_path);
 	g_free (new_path);
 
-	res = sftp_status_to_vfs_result (iobuf_read_result (conn->in_fd, id));
+	res = iobuf_read_result (conn->in_fd, id);
 
 	sftp_connection_unref (conn);
 	sftp_connection_unlock (conn);
@@ -2459,7 +2814,7 @@ do_rename (GnomeVFSMethod  *method,
 	g_free (old_path);
 	g_free (new_path);
 
-	res = sftp_status_to_vfs_result (iobuf_read_result (conn->in_fd, id));
+	res = iobuf_read_result (conn->in_fd, id);
 
 	sftp_connection_unref (conn);
 	sftp_connection_unlock (conn);
@@ -2487,7 +2842,7 @@ do_unlink (GnomeVFSMethod  *method,
 
 	g_free (path);
 
-	res = sftp_status_to_vfs_result (iobuf_read_result (conn->in_fd, id));
+	res = iobuf_read_result (conn->in_fd, id);
 
 	sftp_connection_unref (conn);
 	sftp_connection_unlock (conn);
@@ -2557,7 +2912,7 @@ do_set_file_info (GnomeVFSMethod          *method,
 
 		g_free (path);
 
-		res = sftp_status_to_vfs_result (iobuf_read_result (conn->in_fd, id));
+		res = iobuf_read_result (conn->in_fd, id);
 
 		sftp_connection_unref (conn);
 		sftp_connection_unlock (conn);
@@ -2607,7 +2962,7 @@ do_create_symlink (GnomeVFSMethod   *method,
 
 	g_free (path);
 
-	res = sftp_status_to_vfs_result (iobuf_read_result (conn->in_fd, id));
+	res = iobuf_read_result (conn->in_fd, id);
 
 	sftp_connection_unref (conn);
 	sftp_connection_unlock (conn);
